@@ -4,6 +4,7 @@ import os
 from typing import Any, Awaitable, Callable, Dict, List, Optional, cast
 
 from anthropic import AI_PROMPT, HUMAN_PROMPT, AsyncAnthropic
+from anthropic.types.message import Message as AnthropicMessage
 
 from llmvm.common.container import Container
 from llmvm.common.helpers import Helpers
@@ -22,11 +23,11 @@ class AnthropicExecutor(Executor):
     def __init__(
         self,
         api_key: str,
-        default_model: str = 'claude-3-sonnet-20240229',
+        default_model: str = 'claude-3-5-sonnet-20240620',
         api_endpoint: str = 'https://api.anthropic.com',
         default_max_token_len: int = 200000,
         default_max_output_len: int = 4096,
-        beta: bool = True,
+        max_images: int = 20,
     ):
         super().__init__(
             default_model=default_model,
@@ -35,7 +36,7 @@ class AnthropicExecutor(Executor):
             default_max_output_len=default_max_output_len,
         )
         self.client = AsyncAnthropic(api_key=api_key, base_url=api_endpoint)
-        self.beta = beta
+        self.max_images = max_images
 
     def from_dict(self, message: Dict[str, Any]) -> 'Message':
         role = message['role']
@@ -81,6 +82,9 @@ class AnthropicExecutor(Executor):
         raise ValueError(f'role not found or not supported: {message}')
 
     def wrap_messages(self, messages: List[Message]) -> List[Dict[str, str]]:
+        # todo: this logic is wrong -- if called from execute_direct
+        # it'll unpack the pdf/markdown but not do it properly
+        # as those functions will return multiple messages
         def wrap_message(index: int, content: Content) -> str:
             if isinstance(content, FileContent):
                 return f"<file url={content.url}>{content.get_str()}</file>"
@@ -106,15 +110,15 @@ class AnthropicExecutor(Executor):
 
         messages = [m for m in messages if m.role() != 'system']
 
-        # check to see if there are more than 20 images in the message list
+        # check to see if there are more than self.max_images images in the message list
         image_count = len([m for m in messages if isinstance(m, User) and isinstance(m.message, ImageContent)])
-        if image_count >= 20:
+        if image_count >= self.max_images:
             logging.debug(f'Image count is {image_count}, filtering.')
 
-            # get the top 20 ordered by size, then remove the rest
+            # get the top self.max_images ordered by size, then remove the rest
             images = [m for m in messages if isinstance(m, User) and isinstance(m.message, ImageContent)]
             images.sort(key=lambda x: len(x.message.sequence), reverse=True)
-            smaller_images = images[20:]
+            smaller_images = images[self.max_images:]
 
             # remove the images from the messages list and collapse the previous and next message
             # if there is no other images in between
@@ -152,42 +156,32 @@ class AnthropicExecutor(Executor):
             elif isinstance(messages[i], User) and i < len(messages) - 1:  # is not last message, context messages
                 wrapped.append({'role': 'user', 'content': wrap_message(counter, messages[i].message)})
                 counter += 1
+            elif (
+                isinstance(messages[i], User)
+                and i == len(messages) - 1
+                and (
+                    isinstance(messages[i].message, PdfContent)
+                    or isinstance(messages[i].message, ImageContent)
+                )
+            ):  # is the last message, and it's a pdf or image
+                wrapped.append({'role': 'user', 'content': wrap_message(counter, messages[i].message)})
             elif isinstance(messages[i], User) and i == len(messages) - 1:  # is the last message
                 wrapped.append({'role': 'user', 'content': messages[i].message.get_str()})
             else:
                 wrapped.append({'role': messages[i].role(), 'content': messages[i].message.get_str()})
 
-        messages_list = []
-
-        # for i in range(len(wrapped)):
-        #     if i > 0 and wrapped[i]['role'] == wrapped[i - 1]['role']:
-        #         if wrapped[i]['role'] == 'user':
-        #             messages_list.append({'role': 'assistant', 'content': 'Thanks. Ready for next message.'})
-        #         elif wrapped[i]['role'] == 'assistant':
-        #             messages_list.append({'role': 'user', 'content': 'Thanks. Read for your next message.'})
-        #     messages_list.append(wrapped[i])
-
         return wrapped
 
     def user_token(self):
-        if self.beta:
-            return 'User'
-        else:
-            return HUMAN_PROMPT.replace(':', '').replace('\n', '')
+        return 'User'
 
     def assistant_token(self) -> str:
-        if self.beta:
-            return 'Assistant'
-        else:
-            return AI_PROMPT.replace(':', '').replace('\n', '')
+        return 'Assistant'
 
     def append_token(self) -> str:
-        if self.beta:
-            return ''
-        else:
-            return AI_PROMPT
+        return ''
 
-    def count_tokens(
+    async def count_tokens(
         self,
         messages: List[Message] | List[Dict[str, str]] | str,
         model: Optional[str] = None,
@@ -201,54 +195,44 @@ class AnthropicExecutor(Executor):
             token_count = await self.client.count_tokens(str(content))
             return token_count
 
-        def num_tokens_from_messages(messages):
+        async def num_tokens_from_messages(messages):
             # this is inexact, but it's a reasonable approximation
             num_tokens = 0
             tokens_per_message = 4
             for message in messages:
                 num_tokens += tokens_per_message
                 for _, value in message.items():
-                    num_tokens += asyncio.run(tokenizer_len(value))
+                    num_tokens += await tokenizer_len(value)
             return num_tokens
 
         if isinstance(messages, list) and len(messages) > 0 and isinstance(messages[0], Message):
             dict_messages = self.wrap_messages(cast(List[Message], messages))
-            return num_tokens_from_messages(dict_messages)
+            return await num_tokens_from_messages(dict_messages)
         elif isinstance(messages, list) and len(messages) > 0 and isinstance(messages[0], dict):
-            return num_tokens_from_messages(messages)
+            return await num_tokens_from_messages(messages)
         elif isinstance(messages, str):
-            return num_tokens_from_messages(self.wrap_messages([User(Content(messages))]))
+            return await num_tokens_from_messages(self.wrap_messages([User(Content(messages))]))
         else:
             raise ValueError('cannot calculate tokens for messages: {}'.format(messages))
 
     def name(self) -> str:
         return 'anthropic'
 
-    def __format_prompt(self, messages: List[Dict[str, str]]) -> str:
-        prompt = ''
-        for message in messages:
-            if message['role'] == 'assistant':
-                prompt += f"""{AI_PROMPT} {message['content']}\n\n"""
-            elif message['role'] == 'user':
-                prompt += f"""{HUMAN_PROMPT} {message['content']}\n\n"""
-
-        prompt += f"""{AI_PROMPT}"""
-        return prompt
-
-    async def __aexecute_direct(
+    async def aexecute_direct(
         self,
         messages: List[Dict[str, Any]],
         functions: List[Dict[str, str]] = [],
         model: Optional[str] = None,
-        max_output_tokens: int = 2048,
+        max_output_tokens: int = 4096,
         temperature: float = 0.0,
+        stop_tokens: List[str] = [],
     ) -> TokenStreamManager:
         model = model if model else self.default_model
 
         if functions:
             raise NotImplementedError('functions are not implemented for ClaudeExecutor')
 
-        message_tokens = self.count_tokens(messages=messages, model=model)
+        message_tokens = await self.count_tokens(messages=messages, model=model)
         if message_tokens > self.max_input_tokens(max_output_tokens, model=model):
             raise Exception('Prompt too long. input tokens: {}, output tokens: {}, total: {}, max sontext window: {}'
                             .format(message_tokens,
@@ -270,7 +254,7 @@ class AnthropicExecutor(Executor):
                 logging.warning(f"Removing empty message: {message}")
                 messages.remove(message)
 
-        if Container().get_config_variable('ANTHROPIC_COLLAPSE_MESSAGES', default=False):
+        if Container(throw=False).get_config_variable('ANTHROPIC_COLLAPSE_MESSAGES', default=False):
             collapsed_messages = []
             accumulator = ''
             for i in range(len(messages)):
@@ -303,31 +287,26 @@ class AnthropicExecutor(Executor):
         if messages_list[0]['role'] != 'system' and messages_list[0]['role'] != 'user':
             messages_list.insert(0, {'role': 'user', 'content': 'None.'})
 
-        messages_trace(messages_list)
+        # ugh, anthropic api can't have an assistant message with trailing whitespace...
+        if messages_list[-1]['role'] == 'assistant':
+            messages_list[-1]['content'] = messages_list[-1]['content'].rstrip()
 
-        token_trace = TokenPerf('__aexecute_direct', 'openai', model, prompt_len=message_tokens)  # type: ignore
+        messages_trace([{'role': 'system', 'content': system_message}] + messages_list)
+
+        token_trace = TokenPerf('aexecute_direct', 'anthropic', model)  # type: ignore
         token_trace.start()
 
         try:
-            if self.beta:
-                # AsyncStreamManager[AsyncMessageStream]
-                stream = self.client.messages.stream(
-                    max_tokens=max_output_tokens,
-                    messages=messages_list,  # type: ignore
-                    model=model,
-                    system=system_message,
-                    temperature=temperature,
-                )
-                return TokenStreamManager(stream, token_trace)
-            else:
-                stream = await self.client.completions.create(
-                    max_tokens_to_sample=max_output_tokens,
-                    model=model,
-                    stream=True,
-                    temperature=temperature,
-                    prompt=self.__format_prompt(messages_list),
-                )
-                return TokenStreamManager(stream, token_trace)
+            # AsyncStreamManager[AsyncMessageStream]
+            stream = self.client.messages.stream(
+                max_tokens=max_output_tokens,
+                messages=messages_list,  # type: ignore
+                model=model,
+                system=system_message,
+                temperature=temperature,
+                stop_sequences=stop_tokens,
+            )
+            return TokenStreamManager(stream, token_trace)
         except Exception as e:
             logging.error(e)
             raise
@@ -335,10 +314,12 @@ class AnthropicExecutor(Executor):
     async def aexecute(
         self,
         messages: List[Message],
-        max_output_tokens: int = 2048,
+        max_output_tokens: int = 4096,
         temperature: float = 0.0,
+        stop_tokens: List[str] = [],
         model: Optional[str] = None,
         stream_handler: Callable[[AstNode], Awaitable[None]] = awaitable_none,
+        template_args: Optional[Dict[str, Any]] = None,
     ) -> Assistant:
         model = model if model else self.default_model
 
@@ -354,19 +335,29 @@ class AnthropicExecutor(Executor):
         if not system_message:
             system_message = System(Content('You are a helpful assistant.'))
 
+        expanded_messages = []
+        for message in messages:
+            if isinstance(message, User) and isinstance(message.message, PdfContent):
+                expanded_messages.extend(ObjectTransformers.transform_pdf_content(message.message, self))
+            elif isinstance(message, User) and isinstance(message.message, MarkdownContent):
+                expanded_messages.extend(ObjectTransformers.transform_markdown_content(message.message, self))
+            else:
+                expanded_messages.append(message)
+
         # fresh message list
-        messages_list: List[Dict[str, str]] = self.wrap_messages(messages)
+        messages_list: List[Dict[str, str]] = self.wrap_messages(expanded_messages)
 
         if messages_list[0]['role'] == 'system' and messages_list[1]['role'] != 'user':
             logging.error(f'First message must be from the user after a system prompt: {messages_list}')
         elif messages_list[0]['role'] == 'assistant':
             logging.error(f'First message must be from the user, not assistant: {messages_list}')
 
-        stream = self.__aexecute_direct(
+        stream = self.aexecute_direct(
             messages_list,
             max_output_tokens=max_output_tokens,
             model=model,
             temperature=temperature,
+            stop_tokens=stop_tokens,
         )
 
         text_response = ''
@@ -379,21 +370,17 @@ class AnthropicExecutor(Executor):
             await stream_handler(TokenStopNode())
             perf = stream_async.perf
 
-        if self.beta:
-            _ = await stream_async.get_final_message()  # type: ignore
-        else:
-            async for completion in await stream:  # type: ignore
-                s = completion.completion
-                await stream_handler(Content(s))
-                text_response += s
-            await stream_handler(TokenStopNode())
+        await stream_async.get_final_message()  # this forces an update to the perf object
+        perf.log()
 
         messages_list.append({'role': 'assistant', 'content': text_response})
         conversation: List[Message] = [self.from_dict(m) for m in messages_list]
 
         assistant = Assistant(
             message=conversation[-1].message,
-            messages_context=conversation
+            messages_context=conversation,
+            stop_reason=perf.stop_reason,
+            stop_token=perf.stop_token,
         )
         assistant.perf_trace = perf
         if assistant.message.get_str() == '':
@@ -404,13 +391,15 @@ class AnthropicExecutor(Executor):
     def execute(
         self,
         messages: List[Message],
-        max_output_tokens: int = 2048,
+        max_output_tokens: int = 4096,
         temperature: float = 0.0,
+        stop_tokens: List[str] = [],
         model: Optional[str] = None,
         stream_handler: Optional[Callable[[AstNode], None]] = None,
+        template_args: Optional[Dict[str, Any]] = None,
     ) -> Assistant:
         async def stream_pipe(node: AstNode):
             if stream_handler:
                 stream_handler(node)
 
-        return asyncio.run(self.aexecute(messages, max_output_tokens, temperature, model, stream_pipe))
+        return asyncio.run(self.aexecute(messages, max_output_tokens, temperature, stop_tokens, model, stream_pipe, template_args))
