@@ -1,3 +1,4 @@
+import ast
 import asyncio
 import base64
 import datetime as dt
@@ -9,6 +10,9 @@ import itertools
 import math
 import os
 import re
+import subprocess
+import traceback
+import dateparser
 import typing
 from collections import Counter
 from enum import Enum, IntEnum
@@ -16,8 +20,9 @@ from functools import reduce
 from importlib import resources
 from itertools import cycle, islice
 from logging import Logger
-from typing import Any, Callable, Dict, Generator, List, Optional
-from urllib.parse import urlparse
+from typing import Any, Callable, Dict, Generator, List, Optional, Tuple
+from urllib.parse import urljoin, urlparse
+import zlib
 from zoneinfo import ZoneInfo
 
 import httpx
@@ -27,8 +32,8 @@ from dateutil.relativedelta import relativedelta
 from docstring_parser import parse
 from PIL import Image
 
-from llmvm.common.objects import (Content, ImageContent, MarkdownContent,
-                                  Message, StreamNode, User)
+from llmvm.common.objects import (Content, FunctionCall, ImageContent, MarkdownContent,
+                                  Message, StreamNode, System, User)
 
 
 def write_client_stream(obj):
@@ -50,6 +55,259 @@ def write_client_stream(obj):
 
 
 class Helpers():
+    @staticmethod
+    def apply_unified_diff(original_content, diff_content):
+        original_lines = original_content.splitlines()
+        diff_lines = diff_content.splitlines()
+        modified_lines = original_lines.copy()
+        current_line = 0
+
+        start_line = diff_lines[0]
+        if start_line.startswith("@@") and not start_line.endswith("@@") and "@@" in start_line[2:]:
+            # sonnet botched the format
+            new_start_line = Helpers.in_between_including(start_line, '@@', '@@')
+            next_line = Helpers.after_end(start_line, '@@', '@@').strip()
+            diff_lines.insert(1, next_line)
+            diff_lines[0] = new_start_line
+
+        for i, line in enumerate(diff_lines):
+            if line.startswith("@@"):
+                # @@ -6,6 +6,7 @@ - hunk header
+                # Parse hunk header
+                _, old, new, _ = line.split()
+                old_start = int(old.split(',')[0][1:])
+                current_line = old_start - 1  # -1 because list indices start at 0
+            elif line.startswith("-"):
+                # Remove line
+                if modified_lines[current_line] == line[1:]:
+                    modified_lines.pop(current_line)
+                else:
+                    print(f"Warning: Mismatch at line {current_line + 1}")
+            elif line.startswith("+"):
+                # Add line
+                modified_lines.insert(current_line, line[1:])
+                current_line += 1
+            else:
+                # Context line
+                current_line += 1
+
+        return '\n'.join(modified_lines)
+
+    @staticmethod
+    def apply_context_free_diff(original_content, diff_content):
+        original_lines = original_content.splitlines(True)  # Keep line endings
+        diff_lines = diff_content.splitlines()
+
+        modified_lines = []
+        original_index = 0
+
+        for diff_line in diff_lines:
+            diff_line = diff_line.rstrip('\n')
+            if diff_line.startswith('+'):
+                # Add new line
+                modified_lines.append(diff_line[1:] + '\n')
+            elif diff_line.startswith('-'):
+                # Skip the line in the original content (effectively deleting it)
+                if original_index < len(original_lines):
+                    original_index += 1
+            else:
+                # Context line or unchanged line
+                if original_index < len(original_lines):
+                    if diff_line == original_lines[original_index].rstrip('\n'):
+                        modified_lines.append(original_lines[original_index])
+                        original_index += 1
+                    else:
+                        # Context mismatch - could indicate an error in the diff or original content
+                        modified_lines.append(original_lines[original_index])
+                        original_index += 1
+
+        # Add any remaining lines from the original content
+        modified_lines.extend(original_lines[original_index:])
+
+        # Join the lines back into a single string
+        return ''.join(modified_lines)
+
+    @staticmethod
+    def is_markdown(text):
+        # Define regex patterns for common markdown elements
+        patterns = [
+            r'^\s{0,3}#{1,6}\s',  # Headers
+            r'^\s{0,3}>\s',       # Blockquotes
+            r'^\s{0,3}[-*+]\s',   # Unordered lists
+            r'^\s{0,3}\d+\.\s',   # Ordered lists
+            r'\[.*?\]\(.*?\)',    # Links
+            r'!\[.*?\]\(.*?\)',   # Images
+            r'`.*?`',             # Inline code
+            r'^\s{0,3}```',       # Code blocks
+        ]
+        # Check if any pattern matches the text
+        for pattern in patterns:
+            if re.search(pattern, text, re.MULTILINE):
+                return True
+        return False
+
+    @staticmethod
+    def get_full_url(base_url: str, href: str) -> str:
+        # Parse the base URL to extract scheme and domain
+        parsed_base = urlparse(base_url)
+        base_domain = f"{parsed_base.scheme}://{parsed_base.netloc}"
+
+        # If href is already a full URL, return it
+        if href.startswith(('http://', 'https://')):
+            return href
+
+        # If href starts with '/', join it with the base domain
+        if href.startswith('/'):
+            return urljoin(base_domain, href)
+
+        # For relative URLs, join with the full base URL
+        return str(urljoin(base_url, href))
+
+    @staticmethod
+    def command_substitution(input_string):
+        def execute_command(match):
+            command = match.group(1)
+            try:
+                # Recursively handle nested substitutions
+                command = Helpers.command_substitution(command)
+                result = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)
+                return result.stdout.strip()
+            except subprocess.CalledProcessError as e:
+                return f"Error: {e}"
+
+        pattern = r'\$\(((?:[^()]+|\((?:[^()]+|\([^()]*\))*\))*)\)'
+        return re.sub(pattern, execute_command, input_string)
+
+    @staticmethod
+    def get_callsite(call_str: str, agents: List[Callable]) -> Optional[FunctionCall]:
+        def __get_callsite_helper(
+            call: str,
+            functions: List[Callable]
+        ) -> Optional[Tuple[Callable, Dict[str, Any]]]:
+            function_description: Dict[str, Any] = {}
+
+            if call.startswith('def '):
+                call = call[4:]
+
+            function_name = Helpers.in_between(call, '', '(')
+            if ' ' in function_name or ',' in function_name:
+                return None
+
+            function_arg_str = Helpers.in_between(call, '(', ')')
+            function_args = []
+
+            is_str = False
+            token = ''
+            for i in range(0, len(function_arg_str)):
+                c = function_arg_str[i]
+                if c == '"' and not is_str:
+                    is_str = True
+                    token += c
+                elif c == '"' and is_str:
+                    is_str = False
+                    token += c
+                elif not is_str and c == ',':
+                    function_args.append(token.strip())
+                    token = ''
+                elif not is_str and c == ' ':  # ignore spaces
+                    continue
+                else:
+                    token += c
+
+            if token:
+                function_args.append(token.strip())
+
+            # function_args = [p.strip() for p in Helpers.in_between(call, '(', ')').split(',')]
+            func = functions[0]
+
+            for f in functions:
+                if f.__name__.lower() in function_name.lower():
+                    function_description = Helpers.get_function_description(
+                        f,
+                        openai_format=True
+                    )
+                    func = f
+                    break
+
+            if not function_description:
+                return None
+
+            argument_count = 0
+
+            for _, parameter in function_description['parameters']['properties'].items():
+                if argument_count < len(function_args):
+                    parameter.update({'argument': function_args[argument_count]})
+                argument_count += 1
+
+            return func, function_description
+
+        callsite = __get_callsite_helper(call_str, agents)
+        if callsite:
+            func, function_description = callsite
+            name = function_description['name']
+            arguments = []
+            types = []
+            for arg_name, metadata in function_description['parameters']['properties'].items():
+                # todo if we don't have an argument here, we should ensure that
+                # the function has a default value for the parameter
+                if 'argument' in metadata:
+                    arguments.append({arg_name: metadata['argument']})
+                    types.append({arg_name: metadata['type']})
+
+            return FunctionCall(
+                name=name,
+                args=arguments,
+                types=types,
+                func=func,
+            )
+        return None
+
+    @staticmethod
+    def compare_ast(node1, node2):
+        if type(node1) is not type(node2):
+            return False
+        if isinstance(node1, ast.AST):
+            for k, v in vars(node1).items():
+                if k in ('lineno', 'col_offset', 'ctx'):
+                    continue
+                if not Helpers.compare_ast(v, getattr(node2, k)):
+                    return False
+            return True
+        elif isinstance(node1, list):
+            return len(node1) == len(node2) and all(Helpers.compare_ast(n1, n2) for n1, n2 in zip(node1, node2))
+        else:
+            return node1 == node2
+
+    @staticmethod
+    def compare_code_blocks(code1: str, code2: str):
+        tree1 = ast.parse(code1)
+        tree2 = ast.parse(code2)
+        return Helpers.compare_ast(tree1, tree2)
+
+    @staticmethod
+    def extract_stacktrace_until(stacktrace: str, cls: type):
+        filename = inspect.getfile(cls)
+        lines = stacktrace.split('\n')
+
+        # Find the last occurrence of the filename in the stacktrace
+        file_lines = [i for i, line in enumerate(lines) if filename in line]
+        if not file_lines:
+            return stacktrace
+
+        start_index = file_lines[-1] + 1  # Start from the line after the last occurrence
+        return '\n'.join(lines[start_index:])
+
+    @staticmethod
+    def remove_duplicates(lst, key_func=lambda x: x):
+        seen = set()
+        result = list(filter(lambda x: key_func(x) not in seen and not seen.add(key_func(x)), lst))
+        try:
+            # check to see if any of the strings in the list are substrings of another list item, and if so, remove that
+            sub_dups = [a for a in result if not any(key_func(a) in key_func(b) for b in result if a != b)]
+            return sub_dups
+        except Exception:
+            return result
+
     @staticmethod
     def anthropic_image_tok_count(base64_encoded: str):
         # go from base64 encoded to bytes
@@ -156,6 +414,79 @@ class Helpers():
         markdown_content: MarkdownContent,
         min_width: int,
         min_height: int
+    ) -> List[Content]:
+        pattern = r'(.*?)!\[(?:.*?)\]\((.*?)\)|(.+?)$'
+        chunks = []
+        tasks = []
+        last_end = 0
+        content = markdown_content.get_str()
+        content_list: List[Content] = []
+        idx = 0
+
+        for match in re.finditer(pattern, content, re.DOTALL):
+            before, url, after = match.groups()
+
+            if before and before != content[last_end:match.start()] and last_end != match.start():
+                content_list.append(Content(content[last_end:match.start()]))
+                idx += 1
+
+            if before:
+                content_list.append(Content(before + '![]('))
+                idx += 1
+
+            if url:
+                task = asyncio.create_task(
+                    Helpers.get_image_fuzzy_url(logging, markdown_content.url, url, min_width, min_height)
+                )
+                tasks.append((idx, task, url))
+                content_list.append(Content())
+                idx += 1
+
+            if after:
+                content_list.append(Content(after))
+                idx += 1
+
+            last_end = match.end()
+
+        if last_end < len(content):
+            content_list.append(Content(content[last_end:]))
+            idx += 1
+
+        for idx, task, image_url in tasks:
+            image_bytes = await task
+            if image_bytes:
+                content_list[idx] = ImageContent(image_bytes, image_url)  # type: ignore
+
+        # we're going to collapse the content list here. if there is two text items in a row, we're going to combine them
+        # first, drop all empty content
+        collapsed_content_list = [c for c in content_list if c.sequence]
+        combined_content_list = []
+        current_text = ""
+
+        for c in collapsed_content_list:
+            if not isinstance(c, ImageContent):
+                # If it's a Content instance, add its text to current_text
+                current_text += c.get_str()
+            else:
+                # If current_text has accumulated content, add it to the list as a new Content instance
+                if current_text:
+                    combined_content_list.append(Content(current_text))
+                    current_text = ""  # Reset current_text for the next batch of text
+                # Add the non-text content directly to the list
+                combined_content_list.append(c)
+
+        # If there's any remaining text in current_text, add it to the list as a new Content instance
+        if current_text:
+            combined_content_list.append(Content(current_text))
+
+        return combined_content_list
+
+    @staticmethod
+    async def markdown_content_to_messages_deprecated(
+        logging,
+        markdown_content: MarkdownContent,
+        min_width: int,
+        min_height: int
     ) -> List[Content]:
         markdown_str = markdown_content.get_str()
         tasks = []
@@ -181,9 +512,12 @@ class Helpers():
 
         for idx, task, image_url in tasks:
             image_bytes = await task
-            content_list[idx] = ImageContent(image_bytes, image_url)  # type: ignore
+            if image_bytes:
+                content_list[idx] = ImageContent(image_bytes, image_url)  # type: ignore
 
-        return content_list
+        # should we be collapsing the content list here?
+        results = [c for c in content_list if c.sequence]
+        return results
 
     @staticmethod
     def last_day_of_quarter(year, quarter):
@@ -213,7 +547,7 @@ class Helpers():
         parts = relative_expression.split()
 
         if len(parts) != 2:
-            return parse(relative_expression)  # type: ignore
+            return dateparser.parse(relative_expression)  # type: ignore
 
         value = int(parts[0])
         unit = parts[1].lower()
@@ -227,7 +561,7 @@ class Helpers():
         elif unit == "hours":
             return dt.datetime.now(tz) + dt.timedelta(hours=value)
         else:
-            return parse(relative_expression)  # type: ignore
+            return dateparser.parse(relative_expression)  # type: ignore
 
     @staticmethod
     def load_resize_save(raw_data: bytes, output_format='PNG', max_size=5 * 1024 * 1024) -> bytes:
@@ -283,7 +617,7 @@ class Helpers():
         lineno = exc_traceback.tb_lineno
         filename = frame.f_code.co_filename
 
-        log_message = f"Exception: {e} at line {lineno} in {filename}"
+        log_message = traceback.format_exception(type(e), e, e.__traceback__)
         if message:
             log_message += f": {message}"
 
@@ -404,6 +738,16 @@ class Helpers():
         except Exception:
             return False
 
+    @staticmethod
+    def decompress_if_compressed(byte_stream):
+        try:
+            # Attempt to decompress to see if it's zlib compressed
+            decompressed_data = zlib.decompress(byte_stream)
+            return True, decompressed_data
+        except zlib.error as e:
+            # If decompression fails, it's likely not zlib compressed or the data is corrupted/incomplete
+            return False, byte_stream
+
     @staticmethod
     def image_size(byte_stream: bytes) -> tuple[int, int]:
         try:
@@ -588,6 +932,22 @@ class Helpers():
             return start + ident + end
         return ''
 
+    @staticmethod
+    def after_end(s: str, start: str, end: str) -> str:
+        # Find the start position
+        start_pos = s.find(start)
+        if start_pos == -1:
+            return s
+
+        # Find the end position, starting from after the start token
+        end_pos = s.find(end, start_pos + len(start))
+        if end_pos == -1:
+            return s
+
+        # Extract the content after the end token
+        result = s[end_pos + len(end):]
+        return result
+
     @staticmethod
     def in_between(s, start, end):
         if end == '\n' and '\n' not in s:
@@ -597,6 +957,29 @@ class Helpers():
         part = after_start[:after_start.find(end)]
         return part
 
+    @staticmethod
+    def in_between_including(s, start, end):
+        start_index = s.find(start)
+        if start_index == -1:
+            return ""  # Return empty string if start is not found
+
+        after_start = s[start_index:]
+        end_index = after_start.find(end, len(start))  # Start searching for `end` after `start`
+
+        if end_index == -1:
+            return after_start  # Return everything after start if end is not found
+
+        return s[start_index:start_index + end_index + len(end)]
+
+    @staticmethod
+    def outside_of(s, start, end):
+        if end == '\n' and '\n' not in s:
+            return s[:s.find(start)]
+
+        before_start = s[:s.find(start)]
+        after_end = s[s.find(end) + len(end):]
+        return before_start + after_end
+
     @staticmethod
     def in_between_ends(s, start, end_strs: List[str]):
         # get the text from s between start and any of the end_strs strings.
@@ -660,23 +1043,23 @@ class Helpers():
         return (first, rest[rest.find(end) + len(end):])
 
     @staticmethod
-    def first(predicate, iterable):
+    def first(predicate, iterable, default=None):
         try:
             result = next(x for x in iterable if predicate(x))
             return result
         except StopIteration as ex:
-            return None
+            return default
 
     @staticmethod
     def filter(predicate, iterable):
         return [x for x in iterable if predicate(x)]
 
     @staticmethod
-    def last(predicate, iterable):
+    def last(predicate, iterable, default=None):
         result = [x for x in iterable if predicate(x)]
         if result:
             return result[-1]
-        return None
+        return default
 
     @staticmethod
     def resize_image(screenshot_data, base_width=500):
@@ -793,7 +1176,6 @@ class Helpers():
         if sections:
             results = Helpers.find_closest_sections(query, sections)
         else:
-            print('No sections found')
             return []
         return [a['text'] for a in results]
 
@@ -905,7 +1287,7 @@ class Helpers():
             elif t is float:
                 return 'number'
             else:
-                return 'string'
+                return 'object'
 
         import inspect
 
@@ -919,86 +1301,74 @@ class Helpers():
         func_class = Helpers.__get_class_of_func(func)
         invoked_by = f'{func_class.__name__}.{func_name}' if func_class else func_name
 
-        if openai_format:
-            params = {}
+        params = {}
 
-            for p in inspect.signature(func).parameters:
-                param = inspect.signature(func).parameters[p]
-                parameter = {
-                    param.name: {
-                        'type': parse_type(param.annotation) if param.annotation is not inspect._empty else 'string',
-                        'description': '',
-                    }
+        for p in inspect.signature(func).parameters:
+            param = inspect.signature(func).parameters[p]
+            parameter = {
+                param.name: {
+                    'type': parse_type(param.annotation) if param.annotation is not inspect._empty else 'object',
+                    'description': '',
                 }
-
-                if param.annotation and issubclass(param.annotation, Enum):
-                    values = [v.value for v in param.annotation.__members__.values()]
-                    parameter[param.name]['enum'] = values  # type: ignore
-
-                params.update(parameter)
-
-                # if it's got doc comments, use those instead
-                for p in parse(func.__doc__).params:  # type: ignore
-                    params.update({
-                        p.arg_name: {  # type: ignore
-                            'type': parse_type(p.type_name) if p.type_name is not None else 'string',  # type: ignore
-                            'description': p.description,  # type: ignore
-                        }  # type: ignore
-                    })
-
-            def required_params(func):
-                parameters = inspect.signature(func).parameters
-                return [
-                    name for name, param in parameters.items()
-                    if param.default == inspect.Parameter.empty and param.kind != param.VAR_KEYWORD
-                ]
-
-            function = {
-                'name': invoked_by,
-                'description': description,
-                'parameters': {
-                    'type': 'object',
-                    'properties': params
-                },
-                'required': required_params(func),
             }
+
+            if param.annotation and isinstance(param.annotation, type) and issubclass(param.annotation, Enum):
+                values = [v.value for v in param.annotation.__members__.values()]
+                parameter[param.name]['enum'] = values  # type: ignore
+
+            params.update(parameter)
+
+            # if it's got doc comments, use those instead
+            for p in parse(func.__doc__).params:  # type: ignore
+                params.update({
+                    p.arg_name: {  # type: ignore
+                        'type': parse_type(p.type_name) if p.type_name is not None else 'string',  # type: ignore
+                        'description': p.description,  # type: ignore
+                    }  # type: ignore
+                })
+
+        def required_params(func):
+            parameters = inspect.signature(func).parameters
+            return [
+                name for name, param in parameters.items()
+                if param.default == inspect.Parameter.empty and param.kind != param.VAR_KEYWORD
+            ]
+
+        function = {
+            'name': invoked_by,
+            'description': description,
+            'parameters': {
+                'type': 'object',
+                'properties': params
+            },
+            'required': required_params(func),
+        }
+
+        if openai_format:
             return function
         else:
-            # check to see if parameters are specified in the __doc__, often they're not
-            if not parse(func.__doc__).params:
-                parameters = list(inspect.signature(func).parameters.keys())
-            else:
-                parameters = [p.arg_name for p in parse(func.__doc__).params]
-
-            types = [p.__name__ for p in typing.get_type_hints(func).values()]
-            return_type = typing.get_type_hints(func).get('return')
-
             return {
-                # todo: refactor this to be name
                 'invoked_by': invoked_by,
                 'description': description,
-                'parameters': parameters,
-                'types': types,
-                'return_type': return_type,
+                'parameters': list(params.keys()),
+                'types': [p['type'] for p in params.values()],
+                'return_type': typing.get_type_hints(func).get('return')
             }
 
+    @staticmethod
+    def get_function_description_simple(function: Callable) -> str:
+        description = Helpers.get_function_description(function, openai_format=False)
+        return (f'{description["invoked_by"]}({", ".join(description["parameters"])})  # {description["description"] or "No docstring"}')
+
     @staticmethod
     def get_function_description_flat(function: Callable) -> str:
-        description = Helpers.get_function_description(function, openai_format=False)
-        return (f'{description["invoked_by"]}({", ".join(description["parameters"])})  # {description["description"]}')
-
-    @staticmethod
-    def get_function_description_flat_extra(function: Callable) -> str:
         description = Helpers.get_function_description(function, openai_format=False)
         parameter_type_list = [f"{param}: {typ}" for param, typ in zip(description['parameters'], description['types'])]
-        return_type = description['return_type'].__name__ if description['return_type'] else ''
-        return (f'def {description["invoked_by"]}({", ".join(parameter_type_list)}) -> {return_type}  # {description["description"]}')  # noqa: E501
+        return_type = description['return_type'].__name__ if description['return_type'] else 'Any'
+        return (f'def {description["invoked_by"]}({", ".join(parameter_type_list)}) -> {return_type}  # {description["description"] or "No docstring"}')  # noqa: E501
 
     @staticmethod
-    def load_prompt(prompt_name: str, module: str = 'llmvm.server.prompts.starlark') -> Dict[str, Any]:
-        if not prompt_name.endswith('.prompt'):
-            prompt_name += '.prompt'
-
+    def load_resources_prompt(prompt_name: str, module: str = 'llmvm.server.prompts.python') -> Dict[str, Any]:
         prompt_file = resources.files(module) / prompt_name
 
         with open(prompt_file, 'r') as f:  # type: ignore
@@ -1010,7 +1380,7 @@ class Helpers():
             if '[user_message]' not in prompt:
                 raise ValueError('Prompt file must contain [user_message]')
 
-            system_message = Helpers.in_between(prompt, '[system_message]', '[user_message]')
+            system_message = Helpers.in_between(prompt, '[system_message]', '[user_message]').strip()
             user_message = prompt[prompt.find('[user_message]') + len('[user_message]'):].strip()
             templates = []
 
@@ -1026,15 +1396,33 @@ class Helpers():
             }
 
     @staticmethod
-    def load_and_populate_prompt(
-        prompt_name: str,
+    def get_prompts(
+        prompt_text: str,
         template: Dict[str, str],
         user_token: str = 'User',
         assistant_token: str = 'Assistant',
         append_token: str = '',
-        module: str = 'llmvm.server.prompts.starlark'
-    ) -> Dict[str, Any]:
-        prompt: Dict[str, Any] = Helpers.load_prompt(prompt_name, module)
+    ) -> Tuple[System, User]:
+        if '[system_message]' not in prompt_text:
+            raise ValueError('Prompt file must contain [system_message]')
+
+        if '[user_message]' not in prompt_text:
+            raise ValueError('Prompt file must contain [user_message]')
+
+        system_message = Helpers.in_between(prompt_text, '[system_message]', '[user_message]').strip()
+        user_message = prompt_text[prompt_text.find('[user_message]') + len('[user_message]'):].strip()
+        templates = []
+
+        temp_prompt = prompt_text
+        while '{{' and '}}' in temp_prompt:
+            templates.append(Helpers.in_between(temp_prompt, '{{', '}}'))
+            temp_prompt = temp_prompt.split('}}', 1)[-1]
+
+        prompt = {
+            'system_message': system_message,
+            'user_message': user_message,
+            'templates': templates
+        }
 
         if not template.get('user_token'):
             template['user_token'] = user_token
@@ -1047,10 +1435,95 @@ class Helpers():
             prompt['system_message'] = prompt['system_message'].replace('{{' + key + '}}', value)
             prompt['user_message'] = prompt['user_message'].replace('{{' + key + '}}', value)
 
+        # deal with exec() statements to inject things like datetime
+        import datetime
+        for message_key in ['system_message', 'user_message']:
+            message = prompt[message_key]
+            while '{{' in message and '}}' in message:
+                start = message.find('{{')
+                end = message.find('}}', start)
+                if end == -1:  # No closing '}}' found
+                    break
+
+                key = message[start+2:end]
+                replacement = ''
+
+                if key.startswith('exec('):
+                    try:
+                        replacement = str(eval(key[5:-1]))
+                    except Exception as e:
+                        pass
+                else:
+                    replacement = key
+
+                message = message[:start] + replacement + message[end+2:]
+
+            prompt[message_key] = message
+        return (System(Content(prompt['system_message'])), User(Content(prompt['user_message'] + append_token)))
+
+    @staticmethod
+    def load_and_populate_prompt(
+        prompt_name: str,
+        template: Dict[str, str],
+        user_token: str = 'User',
+        assistant_token: str = 'Assistant',
+        append_token: str = '',
+        module: str = 'llmvm.server.prompts.python'
+    ) -> Dict[str, Any]:
+        prompt: Dict[str, Any] = Helpers.load_resources_prompt(prompt_name, module)
+
+        if not template.get('user_token'):
+            template['user_token'] = user_token
+            template['user_colon_token'] = user_token + ':'
+        if not template.get('assistant_token'):
+            template['assistant_token'] = assistant_token
+            template['assistant_colon_token'] = assistant_token + ':'
+
+        for key, value in template.items():
+            prompt['system_message'] = prompt['system_message'].replace('{{' + key + '}}', value)
+            prompt['user_message'] = prompt['user_message'].replace('{{' + key + '}}', value)
+
+        # deal with exec() statements to inject things like datetime
+        import datetime
+        for message_key in ['system_message', 'user_message']:
+            message = prompt[message_key]
+            while '{{' in message and '}}' in message:
+                start = message.find('{{')
+                end = message.find('}}', start)
+                if end == -1:  # No closing '}}' found
+                    break
+
+                key = message[start+2:end]
+                replacement = ''
+
+                if key.startswith('exec('):
+                    try:
+                        replacement = str(eval(key[5:-1]))
+                    except Exception as e:
+                        pass
+                else:
+                    replacement = key
+
+                message = message[:start] + replacement + message[end+2:]
+
+            prompt[message_key] = message
+
         prompt['user_message'] += f'{append_token}'
         prompt['prompt_name'] = prompt_name
         return prompt
 
+    @staticmethod
+    def populate_prompts(
+        prompt_str: str,
+        template: Dict[str, str],
+        user_token: str = 'User',
+        assistant_token: str = 'Assistant',
+        append_token: str = '',
+        module: str = 'llmvm.server.prompts.python'
+    ) -> Tuple[System, User]:
+        prompt = Helpers.load_and_populate_prompt(prompt_str, template, user_token, assistant_token, append_token, module)
+        return (prompt['system_message'], prompt['user_message'])
+
     @staticmethod
     def prompt_message(
         prompt_name: str,
@@ -1058,7 +1531,19 @@ class Helpers():
         user_token: str = 'User',
         assistant_token: str = 'Assistant',
         append_token: str = '',
-        module: str = 'llmvm.server.prompts.starlark'
+        module: str = 'llmvm.server.prompts.python'
     ) -> Message:
         prompt = Helpers.load_and_populate_prompt(prompt_name, template, user_token, assistant_token, append_token, module)
         return User(Content(prompt['user_message']))
+
+    @staticmethod
+    def prompts(
+        prompt_name: str,
+        template: Dict[str, str],
+        user_token: str = 'User',
+        assistant_token: str = 'Assistant',
+        append_token: str = '',
+        module: str = 'llmvm.server.prompts.python'
+    ) -> Tuple[System, User]:
+        prompt = Helpers.load_and_populate_prompt(prompt_name, template, user_token, assistant_token, append_token, module)
+        return (System(Content(prompt['system_message'])), User(Content(prompt['user_message'])))
