@@ -7,13 +7,20 @@ import os
 from abc import ABC, abstractmethod
 from enum import Enum
 from importlib import resources
-from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar
+from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar, TypedDict
 
-from pydantic import BaseModel
+import numpy as np
+from pydantic import BaseModel, Field
 
 T = TypeVar('T')
 
 
+class DownloadParams(TypedDict):
+    url: str
+    goal: str
+    search_term: str
+
+
 class TokenPriceCalculator():
     def __init__(
         self,
@@ -133,10 +140,11 @@ class Executor(ABC):
     async def aexecute(
         self,
         messages: List['Message'],
-        max_output_tokens: int = 2048,
-        temperature: float = 1.0,
-        stream_handler: Optional[Callable[['AstNode'], Awaitable[None]]] = None,
+        max_output_tokens: int = 4096,
+        temperature: float = 0.0,
+        stop_tokens: List[str] = [],
         model: Optional[str] = None,
+        stream_handler: Optional[Callable[['AstNode'], Awaitable[None]]] = None,
         template_args: Optional[Dict[str, Any]] = None,
     ) -> 'Assistant':
         pass
@@ -186,8 +194,9 @@ class Executor(ABC):
         messages: List['Message'],
         max_output_tokens: int = 2048,
         temperature: float = 1.0,
-        stream_handler: Optional[Callable[['AstNode'], None]] = None,
+        stop_tokens: List[str] = [],
         model: Optional[str] = None,
+        stream_handler: Optional[Callable[['AstNode'], None]] = None,
         template_args: Optional[Dict[str, Any]] = None,
     ) -> 'Assistant':
         pass
@@ -197,7 +206,7 @@ class Executor(ABC):
         pass
 
     @abstractmethod
-    def count_tokens(
+    async def count_tokens(
         self,
         messages: List['Message'] | str,
         model: Optional[str] = None,
@@ -231,12 +240,39 @@ def coerce_types(a, b):
         except ValueError:
             return False
 
+    def is_float(x):
+        return np.isscalar(x) and isinstance(x, (float, np.floating))
+
+    def is_aware(dt):
+        return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None
+
+    if isinstance(a, FunctionCallMeta):
+        a = a.result()
+
+    if isinstance(b, FunctionCallMeta):
+        b = b.result()
+
     # If either operand is a string and represents a number, convert it
     if isinstance(a, str) and is_number(a):
         a = int(a) if '.' not in a else float(a)
     if isinstance(b, str) and is_number(b):
         b = int(b) if '.' not in b else float(b)
 
+    if isinstance(a, dt.date):
+        a = dt.datetime(a.year, a.month, a.day)
+
+    if isinstance(b, dt.date):
+        b = dt.datetime(b.year, b.month, b.day)
+
+    if isinstance(a, dt.datetime) and isinstance(b, dt.datetime):
+        if is_aware(a) and is_aware(b):
+            return a, b
+        elif not is_aware(a) and not is_aware(b):
+            return a, b
+        else:
+            a = a.replace(tzinfo=None)
+            b = b.replace(tzinfo=None)
+
     # If either operand is a string now, convert both to strings
     if isinstance(a, str) or isinstance(b, str):
         return str(a), str(b)
@@ -245,18 +281,16 @@ def coerce_types(a, b):
     if type(a) is type(b):
         return a, b
 
+    # numpy and python floats
+    if is_float(a) and is_float(b):
+        return float(a), float(b)  # type: ignore
+
     # If one is a float and the other an int, convert the int to float
     if isinstance(a, float) and isinstance(b, int):
         return a, float(b)
     if isinstance(b, float) and isinstance(a, int):
         return float(a), b
 
-    if isinstance(a, dt.date):
-        a = dt.datetime(a.year, a.month, a.day)
-
-    if isinstance(b, dt.date):
-        b = dt.datetime(b.year, b.month, b.day)
-
     if isinstance(a, dt.datetime) and isinstance(b, dt.timedelta):
         return a, b
     if isinstance(b, dt.datetime) and isinstance(a, dt.timedelta):
@@ -292,6 +326,7 @@ class LLMCall():
         max_prompt_len: int,
         completion_tokens_len: int,
         prompt_name: str,
+        stop_tokens: List[str] = [],
         stream_handler: Callable[['AstNode'], Awaitable[None]] = awaitable_none
     ):
         self.user_message = user_message
@@ -302,6 +337,7 @@ class LLMCall():
         self.max_prompt_len = max_prompt_len
         self.completion_tokens_len = completion_tokens_len
         self.prompt_name = prompt_name
+        self.stop_tokens = stop_tokens
         self.stream_handler = stream_handler
 
     def copy(self):
@@ -314,6 +350,7 @@ class LLMCall():
             max_prompt_len=self.max_prompt_len,
             completion_tokens_len=self.completion_tokens_len,
             prompt_name=self.prompt_name,
+            stop_tokens=self.stop_tokens,
             stream_handler=self.stream_handler,
         )
 
@@ -465,7 +502,7 @@ class Content(AstNode):
             return str(self.sequence)
 
     def __repr__(self):
-        return f'Content({self.sequence})'
+        return f'Content({self.sequence.__repr__()})'
 
     def get_str(self) -> str:
         return self.__str__()
@@ -476,7 +513,7 @@ class Content(AstNode):
         elif isinstance(self.sequence, str):
             return base64.b64encode(self.sequence.encode('utf-8')).decode('utf-8')
         elif isinstance(self.sequence, list) and len(self.sequence) > 0 and isinstance(self.sequence[0], Content):
-            return base64.b64encode(self.original).decode('utf-8')
+            return base64.b64encode(self.original_sequence).decode('utf-8')  # type: ignore
         else:
             raise ValueError(f'unknown sequence: {self.sequence}')
 
@@ -501,6 +538,22 @@ class ImageContent(Content):
         return f'ImageContent({self.url})'
 
 
+class BrowserContent(Content):
+    def __init__(
+        self,
+        sequence: List[Content],
+        url: str = '',
+    ):
+        super().__init__(sequence, 'browser', url)
+        self.sequence = sequence
+
+    def __str__(self):
+        return f'BrowserContent({self.url})'
+
+    def __repr__(self):
+        return f'BrowserContent({self.url})'
+
+
 class MarkdownContent(Content):
     def __init__(
         self,
@@ -546,7 +599,10 @@ class FileContent(Content):
         self.sequence = sequence
 
     def __str__(self):
-        return f'FileContent({self.url})'
+        return f'FileContent({self.url.__str__()} is_local: {self.is_local()})'
+
+    def __repr__(self):
+        return f'FileContent({self.url.__str__()} is_local: {self.is_local()})'
 
     def is_local(self):
         return os.path.isfile(self.url)
@@ -565,6 +621,7 @@ class Message(AstNode):
         message: Content,
     ):
         self.message: Content = message
+        self.pinned: int = 0  # 0 is not pinned, -1 is pinned last, anything else is pinned
 
     @abstractmethod
     def role(self) -> str:
@@ -686,7 +743,7 @@ class User(Message):
         return str(self.message)
 
     def __repr__(self):
-        return f'Message({self.message})'
+        return f'User({self.message.__repr__()})'
 
 
 class System(Message):
@@ -698,7 +755,7 @@ class System(Message):
             Ask for clarification if a user request is ambiguous.
         ''')
     ):
-        super().__init__(Content(message))
+        super().__init__(message)
 
     def role(self) -> str:
         return 'system'
@@ -707,7 +764,7 @@ class System(Message):
         return str(self.message)
 
     def __repr__(self):
-        return f'SystemPrompt({self.message})'
+        return f'System({self.message.__repr__()})'
 
 
 class Assistant(Message):
@@ -718,12 +775,16 @@ class Assistant(Message):
         messages_context: List[Message] = [],
         system_context: object = None,
         llm_call_context: object = None,
+        stop_reason: str = '',
+        stop_token: str = '',
     ):
         super().__init__(message)
         self.error = error
         self._llm_call_context: object = llm_call_context
         self._system_context = system_context,
         self._messages_context: List[Message] = messages_context
+        self.stop_reason: str = stop_reason
+        self.stop_token: str = stop_token
         self.perf_trace: object = None
 
     def role(self) -> str:
@@ -732,6 +793,9 @@ class Assistant(Message):
     def __str__(self):
         return f'{self.message}'
 
+    def get_str(self):
+        return str(self.message)
+
     def __add__(self, other):
         other_message = str(other)
 
@@ -740,11 +804,16 @@ class Assistant(Message):
             messages_context=self._messages_context,
             system_context=self._system_context,
             llm_call_context=self._llm_call_context,
+            stop_reason=self.stop_reason,
+            stop_token=self.stop_token,
         )
         return assistant
 
     def __repr__(self):
-        return f'Assistant({self.message} {self.error})'
+        if self.error:
+            return f'Assistant({self.message.__repr__()} {self.error})'
+        else:
+            return f'Assistant({self.message.__repr__()})'
 
 
 class Statement(AstNode):
@@ -808,8 +877,21 @@ class FunctionCallMeta(Call):
     def token(self):
         return 'functioncallmeta'
 
+    def __float__(self):
+        return float(self._result)  # type: ignore
+
     def __getattr__(self, name):
-        return getattr(self._result, name)
+        if self._result is not None:
+            return getattr(self._result, name)
+        raise AttributeError(f"'self._result isn't set, and {self.__class__.__name__}' object has no attribute '{name}'")
+
+    def __setstate__(self, state):
+        # Directly set _data without going through __getattr__
+        self._result = state.get('_result')
+
+    def __getstate__(self):
+        # Return a dictionary representing the object's state
+        return {'_result': self._result}
 
     def __str__(self):
         return str(self._result)
@@ -897,7 +979,7 @@ class PandasMeta(Call):
         pandas_df,
     ):
         self.expr_str = expr_str
-        self.df = pandas_df
+        self.df: DataFrame = pandas_df
 
     def result(self) -> object:
         return self._result
@@ -906,11 +988,28 @@ class PandasMeta(Call):
         return 'pandasmeta'
 
     def __str__(self):
-        return str(self.df)
+        str_acc = ''
+        if self.df is not None:
+            str_acc += f'{self.df.info()}\n\n'  # type: ignore
+            str_acc += f'{self.df.describe()()}\n\n'  # type: ignore
+            str_acc += f'{self.df.head()}\n\n'  # type: ignore
+            str_acc += 'head()\n'
+            return str_acc
+        else:
+            return '[]'
 
-    def ask(self, *args, **kwargs) -> object:
-        self._result = self.df.ask(*args, **kwargs)  # type: ignore
-        return self._result
+    def __getattr__(self, name):
+        if name in self.__dict__:
+            return getattr(self.df, name)
+        elif object.__getattribute__(self, 'df') is not None:
+            return getattr(self.df, name)
+        raise AttributeError(f"'self.df isn't set, and {self.__class__.__name__}' object has no attribute '{name}'")
+
+    def __getitem__(self, key):
+        return self.df.__getitem__(key)  # type: ignore
+
+    def __format__(self, format_spec):
+        return format(self.pandas_df, format_spec)
 
 
 class FunctionCall(Call):
@@ -966,11 +1065,13 @@ class Answer(Statement):
         self.error = error
 
     def __str__(self):
-        ret_result = f'Answer({self.result})\n'
-        ret_result = f'Error: {self.error}\n'
-        ret_result += '  Conversation:\n'
-        ret_result += '\n  '.join([str(n) for n in self.conversation])
-        return ret_result
+        if not self.error:
+            return str(self.result())
+        else:
+            return str(self.error)
+
+    def get_str(self):
+        return str(self)
 
     def token(self):
         return 'answer'
@@ -999,8 +1100,11 @@ class SessionThread(BaseModel):
     id: int = -1
     executor: str = ''
     model: str = ''
-    current_mode: str = 'auto'
-    compression: str = 'auto'
+    current_mode: str = ''
+    compression: str = ''
     temperature: float = 0.0
+    stop_tokens: list[str] = []
+    output_token_len: int = 0
     cookies: List[Dict[str, Any]] = []
     messages: List[MessageModel] = []
+    locals_dict: Dict[str, Any] = Field(default_factory=dict, exclude=True)
