@@ -2,18 +2,17 @@ import datetime as dt
 import inspect
 import os
 import time
-from typing import List, cast
+from typing import Any, List, Optional, cast
 
 import openai
 from anthropic import AsyncMessageStream, AsyncMessageStreamManager
 from anthropic import AsyncStream as AnthropicAsyncStream
 from anthropic.types import Completion as AnthropicCompletion
+from anthropic.types import Message as AnthropicMessage
 from google.generativeai.types.generation_types import \
     AsyncGenerateContentResponse as AsyncGeminiStream
 from google.generativeai.types.generation_types import \
     GenerateContentResponse as GeminiCompletion
-from mistralai.models.chat_completion import \
-    ChatCompletionStreamResponse as MistralCompletion
 from openai.types.chat.chat_completion_chunk import \
     ChatCompletionChunk as OAICompletion
 
@@ -46,11 +45,15 @@ class TokenPerf:
         self._start: float = 0.0
         self._stop: float = 0.0
         self._prompt_len: int = prompt_len
+        self._completion_len: int = 0
         self._ticks: List[float] = []
         self.enabled = enabled
         self.log_file = log_file
         self.calculator = TokenPriceCalculator()
         self.request_id = request_id
+        self.stop_reason = ''
+        self.stop_token = ''
+        self.object = None
 
     def start(self):
         if self.enabled:
@@ -70,31 +73,33 @@ class TokenPerf:
             return sum(list) / len(list)
 
         if self.enabled:
-            total_time = self._stop - self._start
-            prompt_time = self._ticks[0] - self._start if self._ticks else 0
-            sample_time = total_time - prompt_time
+            ttlt = self._stop - self._start
+            ttft = self._ticks[0] - self._start if self._ticks else 0
+            completion_time = ttlt - ttft
             try:
-                s_tok_sec = len(self._ticks) / total_time
+                s_tok_sec = len(self._ticks) / ttlt
             except ZeroDivisionError:
                 s_tok_sec = 0.0
             try:
-                p_tok_sec = self._prompt_len / prompt_time
+                p_tok_sec = self._prompt_len / ttft
             except ZeroDivisionError:
                 p_tok_sec = 0.0
             return {
                 'name': self._name,
                 'executor': self._executor,
                 'model': self._model,
-                'total_time': total_time,
-                'prompt_time': prompt_time,
-                'sample_time': sample_time,
+                'ttlt': ttlt,
+                'ttft': ttft,
+                'completion_time': completion_time,
                 'prompt_len': self._prompt_len,
-                'sample_len': len(self._ticks),
+                'completion_len': self._completion_len if self._completion_len > 0 else len(self._ticks),
                 's_tok_sec': s_tok_sec,
                 'p_tok_sec': p_tok_sec,
                 'p_cost': self._prompt_len * self.calculator.input_price(self._model, self._executor),
                 's_cost': len(self._ticks) * self.calculator.output_price(self._model, self._executor),
                 'request_id': self.request_id,
+                'stop_reason': self.stop_reason,
+                'stop_token': self.stop_token,
                 'ticks': self.ticks()
             }
         else:
@@ -113,7 +118,7 @@ class TokenPerf:
     def __str__(self):
         if self.enabled:
             res = self.result()
-            result = f'{dt.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")},{res["name"]},{res["executor"]},{res["model"]},{res["total_time"]},{res["prompt_time"]},{res["sample_time"]},{res["prompt_len"]},{res["sample_len"]},{res["p_tok_sec"]},{res["s_tok_sec"]},{res["request_id"]},{",".join([f"{t:.8f}" for t in res["ticks"]])}'
+            result = f'{dt.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")},{res["name"]},{res["executor"]},{res["model"]},{res["ttlt"]},{res["ttft"]},{res["completion_time"]},{res["prompt_len"]},{res["completion_len"]},{res["p_tok_sec"]},{res["s_tok_sec"]},{res["request_id"]},{res["stop_reason"]},{res["stop_token"]},{",".join([f"{t:.8f}" for t in res["ticks"]])}'
             return result
         else:
             return ''
@@ -124,9 +129,9 @@ class TokenPerf:
             # output \n to the debug stream without using logging.debug
             import sys
             sys.stderr.write('\n')
-            logging.debug(f"total_time: {res['total_time']:.2f} prompt_time: {res['prompt_time']:.2f} sample_time: {res['sample_time']:.2f}")
-            logging.debug(f"prompt_len: {res['prompt_len']} sample_len: {len(res['ticks'])}")
-            logging.debug(f"p_tok_sec: {res['p_tok_sec']:.2f} s_tok_sec: {res['s_tok_sec']:.2f}")
+            logging.debug(f"ttft: {res['ttft']:.2f} ttlt: {res['ttlt']:.2f} completion_time: {res['completion_time']:.2f}")
+            logging.debug(f"prompt_len: {res['prompt_len']} completion_len: {res['completion_len']} model: {res['model']}")
+            logging.debug(f"p_tok_sec: {res['p_tok_sec']:.2f} s_tok_sec: {res['s_tok_sec']:.2f} stop_reason: {res['stop_reason']}")
             logging.debug(f"p_cost: ${res['p_cost']:.5f} s_cost: ${res['s_cost']:.5f} request_id: {res['request_id']}")
 
     def log(self):
@@ -134,7 +139,7 @@ class TokenPerf:
             self.debug()
             if not os.path.exists(os.path.expanduser(self.log_file)):
                 with open(os.path.expanduser(self.log_file), 'w') as f:
-                    f.write('name,executor,model,total_time,prompt_time,prompt_tokens,sample_time,prompt_len,sample_len,p_tok_sec,s_tok_sec,p_cost,s_cost,request_id,ticks\n')
+                    f.write('name,executor,model,ttlt,ttft,prompt_tokens,completion_time,prompt_len,completion_len,p_tok_sec,s_tok_sec,p_cost,s_cost,request_id,stop_reason,stop_token,ticks\n')
             with open(os.path.expanduser(self.log_file), 'a') as f:
                 result = str(self)
                 f.write(result + '\n')
@@ -143,16 +148,18 @@ class TokenPerf:
             return {
                 'name': self._name,
                 'executor': self._executor,
-                'total_time': 0.0,
-                'prompt_time': 0.0,
-                'sample_time': 0.0,
+                'ttlt': 0.0,
+                'ttft': 0.0,
+                'completion_time': 0.0,
                 'prompt_len': 0,
-                'sample_len': 0,
+                'completion_len': 0,
                 'p_tok_sec': 0.0,
                 's_tok_sec': 0.0,
                 'p_cost': 0.0,
                 's_cost': 0.0,
                 'request_id': '',
+                'stop_reason': '',
+                'stop_token': '',
                 'ticks': []
             }
 
@@ -172,8 +179,8 @@ class LoggingAsyncIterator:
             if isinstance(result, AnthropicCompletion):
                 return cast(str, result.completion or '')
             elif isinstance(result, OAICompletion):
-                return cast(str, result.choices[0].delta.content or '')  # type: ignore
-            elif isinstance(result, MistralCompletion):
+                if result.choices[0].finish_reason:
+                    self.perf.stop_reason = result.choices[0].finish_reason
                 return cast(str, result.choices[0].delta.content or '')  # type: ignore
             elif isinstance(result, GeminiCompletion):
                 return cast(str, result.text or '')
@@ -184,7 +191,6 @@ class LoggingAsyncIterator:
         except StopAsyncIteration:
             if self.perf.enabled:
                 self.perf.stop()
-                self.perf.log()
             raise
 
 
@@ -196,6 +202,7 @@ class TokenStreamWrapper:
     ):
         self.stream = original_stream
         self.perf = token_perf
+        self.object: Optional[Any] = None  # used for Anthropic's AsyncMessageStream needed for get_final_message()
 
     def __aiter__(self):
         return LoggingAsyncIterator(self.stream, self.perf)
@@ -209,19 +216,31 @@ class TokenStreamWrapper:
 
     # we're proxying the .text_stream, which doesn't have get_final_message()
     # but we don't need it as we're streaming
-    async def get_final_message(self):
-        return ''
+    async def get_final_message(self) -> Optional[AnthropicMessage]:
+        if isinstance(self.object, AsyncMessageStream):
+            final_message = await self.object.get_final_message()
+            self.perf._prompt_len = final_message.usage.input_tokens
+            self.perf._completion_len = final_message.usage.output_tokens
+            self.perf.object = final_message  # type: ignore
+            self.perf.stop_reason = str(final_message.stop_reason) if final_message.stop_reason else ''
+            self.perf.stop_token = final_message.stop_sequence if final_message.stop_sequence else ''
+            await self.object.close()
+            return final_message
+        else:
+            return None
 
 
 class TokenStreamManager:
     def __init__(
         self,
         stream: AsyncMessageStreamManager | AnthropicAsyncStream[AnthropicCompletion] | openai.AsyncStream,  # type: ignore
-        token_perf: 'TokenPerf'
+        token_perf: 'TokenPerf',
+        stop_tokens: List[str] = []
     ):
         self.stream = stream
         self.perf = token_perf
         self.token_perf_wrapper = None
+        self.stop_tokens = stop_tokens
 
     async def __aenter__(self) -> TokenStreamWrapper:
         self.perf.start()
@@ -234,6 +253,8 @@ class TokenStreamManager:
                 with open(os.path.expanduser(Container().get_config_variable('LLMVM_SHARE') + f'/{self.perf.request_id}.json'), 'w') as f:
                     f.write(result.response._request._content.decode('utf-8'))  # type: ignore
             self.token_perf_wrapper = TokenStreamWrapper(result.text_stream, self.perf)  # type: ignore
+            self.token_perf_wrapper.object = result
+
             return self.token_perf_wrapper
         elif isinstance(self.stream, openai.AsyncStream):
             self.token_perf_wrapper = TokenStreamWrapper(self.stream, self.perf)
