@@ -1,20 +1,23 @@
 import re
-from typing import Any, Callable, Dict, List
+from typing import Any, Callable, Dict, List, TypedDict
 
 from googlesearch import search as google_search
 
 from llmvm.common.container import Container
 from llmvm.common.helpers import Helpers, write_client_stream
 from llmvm.common.logging_helpers import setup_logging
-from llmvm.common.objects import (Content, LLMCall, Message,
+from llmvm.common.objects import (Content, DownloadParams, LLMCall, Message,
                                   TokenCompressionMethod, User, bcl)
-from llmvm.server.starlark_execution_controller import ExecutionController
+from llmvm.server.base_library.content_downloader import WebAndContentDriver
+from llmvm.server.python_execution_controller import ExecutionController
 from llmvm.server.tools.search import SerpAPISearcher
 from llmvm.server.tools.webhelpers import WebHelpers
 from llmvm.server.vector_search import VectorSearch
 
 logging = setup_logging()
 
+
+
 class Searcher():
     def __init__(
         self,
@@ -23,7 +26,7 @@ class Searcher():
         original_code: str,
         original_query: str,
         vector_search: VectorSearch,
-        total_links_to_return: int = 2,
+        total_links_to_return: int = 3,
     ):
         self.query = expr
         self.original_code = original_code
@@ -31,7 +34,12 @@ class Searcher():
         self.controller = controller
         self.query_expansion = 2
 
-        self.parser = WebHelpers.get_url
+            # url: str,
+            # goal: str,
+            # search_term: Optional[str],
+            # controller: ExecutionController,
+
+        self.parser: Callable = lambda x: Content('Searcher().search() parser: no parser found.')
         self.ordered_snippets: List = []
         self.index = 0
         self._result = None
@@ -71,14 +79,13 @@ class Searcher():
 
     def search_research_hook(self, query: str):
         if not Container().get_config_variable('SERPAPI_API_KEY'):
-            return self.search_hook('https://news.google.com/search?q=', query)
+            return self.search_hook('https://www.google.com/search?q=', query)
         else:
-            # likely want more thorough answers so we'll return more results
-            self.total_links_to_return = 10
             return SerpAPISearcher().search_research(query)
 
     def search(
         self,
+        titles_seen: List[str] = [],
     ) -> List[Content]:
         # todo: we should probably return the Search instance, so we can futz with it later on.
         query_expander = self.controller.execute_llm_call(
@@ -131,11 +138,29 @@ class Searcher():
 
         def url_to_text(result: Dict[str, Any]) -> Content:
             if 'link' in result and isinstance(result['link'], Dict) and 'link' in result['link']:
-                return WebHelpers.get_url(result['link']['link'])
+                if Container().get_config_variable('LLMVM_FULL_PROCESSING', default=False):
+                    return WebAndContentDriver().download_with_goal(download = {
+                        'url': str(result['link']['link']),
+                        'goal': self.original_query,
+                        'search_term': self.query
+                        },
+                        controller=self.controller
+                    )
+                else:
+                    return WebHelpers.get_url(result['link']['link'])
             elif 'link' in result:
-                return WebHelpers.get_url(result['link'])  # type: ignore
+                if Container().get_config_variable('LLMVM_FULL_PROCESSING', default=False):
+                    return WebAndContentDriver().download_with_goal(download = {
+                        'url': str(result['link']),
+                        'goal': self.original_query,
+                        'search_term': self.query
+                        },
+                        controller=self.controller
+                    )
+                else:
+                    return WebHelpers.get_url(str(result['link']))
             else:
-                return Content()
+                return Content('Searcher().search() url_to_text: no link found in result, so no content to return.')
 
         def yelp_to_text(reviews: Dict[Any, Any]) -> Content:
             return_str = f"{reviews['title']} in {reviews['neighborhood']}."
@@ -163,7 +188,6 @@ class Searcher():
 
         engines = {
             'Google Search': {'searcher': self.search_google_hook, 'parser': url_to_text, 'description': 'Google Search is a general web search engine that is good at answering questions, finding knowledge and information, and has a complete scan of the Internet.'},  # noqa:E501
-            'Google News': {'searcher': self.search_news_hook, 'parser': url_to_text, 'description': 'Google News Search is a news search engine. This engine is excellent at finding news about particular topics, people, companies and entities.'},  # noqa:E501
             'Google Patent Search': {'searcher': self.search_google_hook, 'parser': url_to_text, 'description': 'Google Patent Search is a search engine that is exceptional at findind matching patents for a given query.'},  # noqa:E501
             'Yelp Search': {'searcher': SerpAPISearcher().search_yelp, 'parser': yelp_to_text, 'description': 'Yelp is a search engine dedicated to finding geographically local establishments, restaurants, stores etc and extracing their user reviews.'},  # noqa:E501
             'Local Files Search': {'searcher': self.vector_search.search, 'parser': local_to_text, 'description': 'Local file search engine. Searches the users hard drive for content in pdf, csv, html, doc and docx files.'},  # noqa:E501
@@ -204,7 +228,7 @@ class Searcher():
                 self.parser = engines[key]['parser']
                 searcher = engines[key]['searcher']
 
-        write_client_stream(Content(f"I'm using the {engine} engine to perform search.\n"))
+        write_client_stream(Content(f"I think the {engine} engine is best to perform a search for {self.query}\n"))
 
         # perform the search
         search_results = []
@@ -267,6 +291,7 @@ class Searcher():
                         'snippets': '\n'.join(
                             [f'* {str(key)}: {value["title"]} {value["snippet"]}' for key, value in snippets.items()]
                         ),
+                        'seen_list': '\n'.join(titles_seen),
                     },
                     user_token=self.controller.get_executor().user_token(),
                     assistant_token=self.controller.get_executor().assistant_token(),
@@ -290,8 +315,8 @@ class Searcher():
         try:
             ranked_results = eval(str(result_rank.message))
         except SyntaxError as ex:
-            logging.debug(f"SyntaxError: {ex}")
-            pattern = r'\[\s*(-?\d+\s*,\s*)*-?\d+\s*\]'
+            logging.debug(f"Searcher.search() SyntaxError: {ex}, trying again with regex list extractor")
+            pattern = r'\[(?:"[^"]*"(?:,\s*)?)+\]'
             match = re.search(pattern, str(result_rank.message))
 
             if match:
