@@ -3,7 +3,7 @@ import os
 import shutil
 import sys
 from importlib import resources
-from typing import List, Optional, cast
+from typing import Callable, List, Optional, cast
 
 import async_timeout
 import jsonpickle
@@ -21,16 +21,15 @@ from llmvm.common.container import Container
 from llmvm.common.gemini_executor import GeminiExecutor
 from llmvm.common.helpers import Helpers
 from llmvm.common.logging_helpers import setup_logging
-from llmvm.common.mistral_executor import MistralExecutor
 from llmvm.common.objects import (Answer, Assistant, AstNode, Content,
-                                  DownloadItem, FileContent, MessageModel,
+                                  DownloadItem, DownloadParams, FileContent, MessageModel,
                                   SessionThread, Statement, StopNode,
                                   TokenCompressionMethod, User,
                                   compression_enum)
 from llmvm.common.openai_executor import OpenAIExecutor
 from llmvm.server.persistent_cache import PersistentCache
-from llmvm.server.starlark_execution_controller import ExecutionController
-from llmvm.server.tools.firefox import FirefoxHelpers
+from llmvm.server.python_execution_controller import ExecutionController
+from llmvm.server.tools.chrome import ChromeHelpers
 from llmvm.server.vector_search import VectorSearch
 from llmvm.server.vector_store import VectorStore
 
@@ -100,8 +99,9 @@ def get_controller(controller: Optional[str] = None) -> ExecutionController:
         anthropic_executor = AnthropicExecutor(
             api_key=os.environ.get('ANTHROPIC_API_KEY', ''),
             default_model=Container().get_config_variable('anthropic_model', 'LLMVM_MODEL'),
-            api_endpoint=Container().get_config_variable('anthropic_api_base', 'LLMVM_API_BASE'),
+            api_endpoint=Container().get_config_variable('anthropic_api_base', 'ANTHROPIC_API_BASE'),
             default_max_token_len=int(Container().get_config_variable('anthropic_max_tokens')),
+            default_max_output_len=int(Container().get_config_variable('anthropic_max_output_tokens')),
         )
         anthropic_controller = ExecutionController(
             executor=anthropic_executor,
@@ -111,25 +111,12 @@ def get_controller(controller: Optional[str] = None) -> ExecutionController:
             continuation_passing_style=False,
         )
         return anthropic_controller
-    elif controller == 'mistral':
-        mistral_executor = MistralExecutor(
-            api_key=os.environ.get('MISTRAL_API_KEY', ''),
-            default_model=Container().get_config_variable('mistral_model', 'LLMVM_MODEL'),
-            default_max_token_len=int(Container().get('mistral_max_tokens')),
-        )
-        mistral_controller = ExecutionController(
-            executor=mistral_executor,
-            agents=agents,  # type: ignore
-            vector_search=vector_search,
-            edit_hook=None,
-            continuation_passing_style=False,
-        )
-        return mistral_controller
     elif controller == 'gemini':
         gemini_executor = GeminiExecutor(
             api_key=os.environ.get('GOOGLE_API_KEY', ''),
             default_model=Container().get_config_variable('gemini_model', 'LLMVM_MODEL'),
             default_max_token_len=int(Container().get('gemini_max_tokens')),
+            default_max_output_len=int(Container().get('gemini_max_output_tokens')),
         )
         gemini_controller = ExecutionController(
             executor=gemini_executor,
@@ -143,8 +130,9 @@ def get_controller(controller: Optional[str] = None) -> ExecutionController:
         openai_executor = OpenAIExecutor(
             api_key=os.environ.get('OPENAI_API_KEY', ''),
             default_model=Container().get_config_variable('openai_model', 'LLMVM_MODEL'),
-            api_endpoint=Container().get_config_variable('openai_api_base', 'LLMVM_API_BASE'),
+            api_endpoint=Container().get_config_variable('openai_api_base', 'OPENAI_API_BASE'),
             default_max_token_len=int(Container().get('openai_max_tokens')),
+            default_max_output_len=int(Container().get('openai_max_output_tokens')),
         )
 
         openai_controller = ExecutionController(
@@ -160,7 +148,7 @@ def get_controller(controller: Optional[str] = None) -> ExecutionController:
 def __get_thread(id: int) -> SessionThread:
     if not cache_session.has_key(id) or id <= 0:
         id = cache_session.gen_key()
-        thread = SessionThread(current_mode='tool', id=id)
+        thread = SessionThread(id=id)
         cache_session.set(thread.id, thread)
     return cast(SessionThread, cache_session.get(id))
 
@@ -265,13 +253,15 @@ async def download(
         async def execute_and_signal():
             stream_handler = callback
             from llmvm.server.base_library.content_downloader import \
-                ContentDownloader
+                WebAndContentDriver
 
             # todo thread cookies through here
-            downloader = ContentDownloader(
-                expr=download_item.url,
-            )
-            content: Content = downloader.download()
+            downloader = WebAndContentDriver()
+            content: Content = downloader.download(download={
+                'url': download_item.url,
+                'goal': '',
+                'search_term': ''
+            })
             queue.put_nowait(StopNode())
 
             if content:
@@ -333,11 +323,11 @@ async def health():
     logging.debug('/health')
     return {'status': 'ok'}
 
-@app.get('/firefox')
-async def firefox(url: str = 'https://9600.dev'):
-    logging.debug(f'/firefox?url={url}')
-    firefox = FirefoxHelpers()
-    result = await firefox.get_url(url)
+@app.get('/browser')
+async def browser(url: str = 'https://9600.dev'):
+    logging.debug(f'/browser?url={url}')
+    chrome = ChromeHelpers()
+    result = await chrome.get_url(url)
     return JSONResponse(content=result)
 
 @app.post('/v1/chat/cookies', response_model=None)
@@ -352,137 +342,16 @@ async def set_cookies(requests: Request):
     cache_session.set(thread.id, thread)
     return thread
 
-@app.post('/v1/chat/code_completions', response_model=None)
-async def code_completions(request: SessionThread):
-    thread = request
-    messages = [MessageModel.to_message(m) for m in thread.messages]  # type: ignore
-    local_files = []
-
-    # todo: refector this crap, it's not good.
-    # typically there are User messages with FileContent Content type in them
-    # we need to download the files, store them locally, then remove them from the messages
-    # as the prompt will generate code that will be able to access the files locally
-
-    # alternatively, we could just pass the files in as a list of strings
-    try:
-        for file_content in [m.message for m in messages if m.message is not None and isinstance(m.message, FileContent)]:
-            # check to see if the FileContent is just a pointer to a local file
-            if file_content.is_local():
-                local_files.append(file_content.url)
-            else:
-                # put it in the CDN directory so the runtime can reference it
-                # file_content.url is absolute path to client.py directory
-                dirname = os.path.dirname(file_content.url)
-                base_name = os.path.basename(file_content.url)
-
-                os.makedirs(f"{cdn_directory}/{thread.id}{dirname}", exist_ok=True)
-                with open(f"{cdn_directory}/{thread.id}{dirname}/{base_name}", "wb") as buffer:
-                    buffer.write(file_content.sequence)  # type: ignore
-                    local_files.append(f"{cdn_directory}/{thread.id}{dirname}/{base_name}")
-
-        # remove the FileContent messages from the messages list
-        messages = [message for message in messages if not isinstance(message.message, FileContent)]
-
-    except Exception as e:
-        raise HTTPException(status_code=500, detail=f"Exception: {e}")
-
-    if not cache_session.has_key(thread.id) or thread.id <= 0:
-        temp = __get_thread(0)
-        thread.id = temp.id
-
-    mode = thread.current_mode
-    compression = compression_enum(thread.compression)
-    queue = asyncio.Queue()
-
-    # set the defaults, or use what the SessionThread thread asks
-    if thread.executor and thread.model:
-        controller = get_controller(thread.executor)
-        model = thread.model if thread.model else controller.get_executor().get_default_model()
-    # either the executor or the model is not set, so use the defaults
-    # and update the thread
-    else:
-        logging.debug('Either the executor or the model is not set. Updating thread.')
-        controller = get_controller()
-        model = controller.get_executor().get_default_model()
-        thread.executor = controller.get_executor().name()
-        thread.model = model
-
-    logging.debug(f'/v1/chat/code_completions?id={thread.id}&mode={mode}&model={model}&executor={thread.executor}')
-
-    if len(messages) == 0:
-        raise HTTPException(status_code=400, detail='No messages provided')
-
-    async def callback(token: AstNode):
-        queue.put_nowait(token)
-
-    async def stream():
-        def handle_exception(task):
-            if not task.cancelled() and task.exception() is not None:
-                Helpers.log_exception(logging, task.exception())
-                queue.put_nowait(StopNode())
-
-        async def execute_and_signal():
-            result = await controller.aexecute(
-                messages=messages,
-                temperature=thread.temperature,
-                mode='code',
-                stream_handler=callback,
-                model=model,
-                compression=compression,
-                template_args={'files': local_files}
-            )
-            queue.put_nowait(StopNode())
-            return result
-
-        task = asyncio.create_task(execute_and_signal())
-        task.add_done_callback(handle_exception)
-
-        while True:
-            data = await queue.get()
-            if isinstance(data, StopNode):
-                break
-            yield data
-
-        try:
-            await task
-        except Exception as e:
-            pass
-
-        # error handling
-        if task.exception() is not None:
-            thread.messages.append(MessageModel.from_message(Assistant(Content(f'Error: {str(task.exception())}'))))
-            yield thread.model_dump()
-            return
-
-        statements: List[Statement] = task.result()
-
-        # todo parse Answers into Assistants for now
-        results = []
-        for statement in statements:
-            if isinstance(statement, Answer):
-                results.append(Assistant(Content(str(cast(Answer, statement).result()))))
-            elif isinstance(statement, Assistant):
-                results.append(statement)
-
-        if len(results) > 0:
-            for result in results:
-                thread.messages.append(MessageModel.from_message(result))
-            cache_session.set(thread.id, thread)
-            yield thread.model_dump()
-        else:
-            # todo need to do something here to deal with error cases
-            yield thread.model_dump()
-
-    return StreamingResponse(stream_response(stream()), media_type='text/event-stream')  # media_type="application/json")
-
-
-@app.post('/v1/chat/tools_completions', response_model=None)
+@app.post('/v1/tools/completions', response_model=None)
 async def tools_completions(request: SessionThread):
     thread = request
 
     if not cache_session.has_key(thread.id) or thread.id == 0:
         temp = __get_thread(0)
         thread.id = temp.id
+    # locals_dict needs to be set, grab it from the cache
+    elif cache_session.has_key(thread.id) and not thread.locals_dict:
+        thread.locals_dict = cache_session.get(thread.id).locals_dict  # type: ignore
 
     messages = [MessageModel.to_message(m) for m in thread.messages]  # type: ignore
     mode = thread.current_mode
@@ -503,8 +372,7 @@ async def tools_completions(request: SessionThread):
         thread.executor = controller.get_executor().name()
         thread.model = model
 
-    logging.debug(f'/v1/chat/tools_completions?id={thread.id}&mode={mode}&model={model} \
-                  &executor={thread.executor}&compression={thread.compression}&cookies={thread.cookies}')
+    logging.debug(f'/v1/tools/completions?id={thread.id}&mode={mode}&model={model}&executor={thread.executor}&compression={thread.compression}&cookies={thread.cookies}&temperature={thread.temperature}')  # NOQA: E501
 
     if len(messages) == 0:
         raise HTTPException(status_code=400, detail='No messages provided')
@@ -523,18 +391,33 @@ async def tools_completions(request: SessionThread):
                 queue.put_nowait(StopNode())
 
         async def execute_and_signal():
-            # todo: this is a hack
-            result = await controller.aexecute(
-                messages=messages,
-                temperature=thread.temperature,
-                mode=mode,
-                stream_handler=callback,
-                model=model,
-                compression=compression,
-                cookies=cookies
-            )
-            queue.put_nowait(StopNode())
-            return result
+            if thread.current_mode == 'direct':
+                result = await controller.aexecute(
+                    messages=messages,
+                    temperature=thread.temperature,
+                    model=model,
+                    mode=thread.current_mode,
+                    compression=compression,
+                    cookies=cookies,
+                    stream_handler=callback,
+                )
+                queue.put_nowait(StopNode())
+                return result
+            else:
+                # todo: this is a hack
+                result, locals_dict = await controller.aexecute_continuation(
+                    messages=messages,
+                    temperature=thread.temperature,
+                    stream_handler=callback,
+                    model=model,
+                    compression=compression,
+                    cookies=cookies,
+                    agents=cast(List[Callable], agents),
+                    locals_dict=thread.locals_dict
+                )
+                queue.put_nowait(StopNode())
+                thread.locals_dict = locals_dict
+                return result
 
         task = asyncio.create_task(execute_and_signal())
         task.add_done_callback(handle_exception)
@@ -561,11 +444,15 @@ async def tools_completions(request: SessionThread):
 
         # todo parse Answers into Assistants for now
         results = []
-        for statement in statements:
-            if isinstance(statement, Answer):
-                results.append(Assistant(Content(str(cast(Answer, statement).result()))))
-            elif isinstance(statement, Assistant):
-                results.append(statement)
+
+        assistant_result = '\n\n'.join([str(statement) for statement in reversed(statements)])
+        results.append(Assistant(Content(assistant_result)))
+
+        # for statement in reversed(statements):
+        #     if isinstance(statement, Answer):
+        #         results.append(Assistant(Content(str(cast(Answer, statement).result()))))
+        #     elif isinstance(statement, Assistant):
+        #         results.append(statement)
 
         if len(results) > 0:
             for result in results:
@@ -591,7 +478,7 @@ if __name__ == '__main__':
     default_model = Container().get_config_variable(default_model_str, 'LLMVM_MODEL', default='')
     rich.print(f'[cyan]Default executor is: {default_controller}[/cyan]')
     rich.print(f'[cyan]Default model is: {default_model}[/cyan]')
-    rich.print(f'[cyan]Make sure to `playwright install firefox`.[/cyan]')
+    rich.print(f'[cyan]Make sure to `playwright install`.[/cyan]')
 
     for agent in agents:
         rich.print(f'[green]Loaded agent: {agent.__name__}[/green]')  # type: ignore
