@@ -2,7 +2,6 @@ import asyncio
 import os
 import re
 import unicodedata
-from typing import Callable, Dict, Generator
 from urllib.parse import urlparse
 
 from bs4 import BeautifulSoup
@@ -10,16 +9,24 @@ from markdownify import MarkdownConverter
 
 from llmvm.common.helpers import write_client_stream
 from llmvm.common.logging_helpers import setup_logging
-from llmvm.common.objects import Content, MarkdownContent
-from llmvm.server.tools.firefox import FirefoxHelpers
+from llmvm.common.objects import Content, DownloadParams, MarkdownContent
+from llmvm.server.tools.chrome import ChromeHelpers
 from llmvm.server.tools.search import SerpAPISearcher
 
 logging = setup_logging()
 
-class IgnoringScriptConverter(MarkdownConverter):
+class VerboseConverter(MarkdownConverter):
     def convert_script(self, el, text, convert_as_inline):
         return ''
 
+    def convert_input(self, el, text, convert_as_inline):
+        return str(el)
+
+    def convert_textarea(self, el, text, convert_as_inline):
+        return str(el)
+
+    def convert_label(self, el, text, convert_as_inline):
+        return str(el)
 
 class WebHelpers():
     @staticmethod
@@ -53,7 +60,9 @@ class WebHelpers():
                 else:
                     lines.append(line)
                     blank_counter = 0
-            return '\n'.join(lines)
+
+            lines = [line.strip() for line in lines]
+            return '\n'.join([line for line in lines if line])
 
         logging.debug(f'WebHelpers.convert_html_to_markdown_soup: {html[:25]}')
         soup = BeautifulSoup(html, features='lxml')
@@ -61,30 +70,33 @@ class WebHelpers():
         for data in soup(['style', 'script']):
             data.decompose()
 
-        result = IgnoringScriptConverter().convert_soup(soup)
+        result = VerboseConverter().convert_soup(soup)
         cleaned_result = clean_markdown(result)
         return MarkdownContent(sequence=unicodedata.normalize('NFKD', cleaned_result), url=url)
 
     @staticmethod
     def get_linkedin_profile(linkedin_url: str) -> str:
+        """
+        Extracts the career information from a person's LinkedIn profile from a given LinkedIn url and returns
+        the career information as a string.
+        """
         from llmvm.common.pdf import PdfHelpers
-        """Extracts the career information from a person's LinkedIn profile from a given LinkedIn url"""
         logging.debug('WebHelpers.get_linkedin_profile: {}'.format(linkedin_url))
 
-        firefox_helpers = FirefoxHelpers()
-        asyncio.run(firefox_helpers.goto(linkedin_url))
-        asyncio.run(firefox_helpers.wait_until_text('Experience'))
-        pdf_file = asyncio.run(firefox_helpers.pdf())
+        chrome_helpers = ChromeHelpers()
+        asyncio.run(chrome_helpers.goto(linkedin_url))
+        asyncio.run(chrome_helpers.wait_until_text('Experience'))
+        pdf_file = asyncio.run(chrome_helpers.pdf())
         data = PdfHelpers.parse_pdf(pdf_file)
         os.remove(pdf_file)
-        asyncio.run(firefox_helpers.close())
+        asyncio.run(chrome_helpers.close())
         return data
 
     @staticmethod
     def search_linkedin_profile(first_name: str, last_name: str, company_name: str) -> str:
         """
         Searches for the LinkedIn profile of a given first name and last name and optional company name and returns the
-        LinkedIn profile. If you use this method you do not need to call get_linkedin_profile.
+        LinkedIn profile information as a string. If you call this method you do not need to call get_linkedin_profile().
         """
         searcher = SerpAPISearcher()
         links = searcher.search_internet('{} {}, {} linkedin profile site:linkedin.com/in/'.format(
@@ -113,7 +125,8 @@ class WebHelpers():
         """
         # todo this is redundant with ContentDownloader
         from llmvm.server.base_library.content_downloader import \
-            ContentDownloader
+            WebAndContentDriver
         logging.debug('WebHelpers.get_url: {}'.format(url))
-        downloader = ContentDownloader(url)
-        return downloader.download()
+        downloader = WebAndContentDriver()
+        download_params: DownloadParams = {'url': url, 'goal': '', 'search_term': ''}
+        return downloader.download(download_params)
